typedef struct vectorStruct {
	double x,y;

	inline vectorStruct ();
	inline vectorStruct (double x, double y);

	inline vectorStruct &operator+=(vectorStruct const &v);
	inline vectorStruct &operator-=(vectorStruct const &v);
	inline int operator==(vectorStruct const &v);
	inline int operator<(vectorStruct const &v);
	inline int operator>(vectorStruct const &v);
} VECTOR;

#define	INVALID_VECTOR	VECTOR(-65535.0,-65535.0)

inline VECTOR operator-(VECTOR const &v1, VECTOR const &v2)
{
	return VECTOR(v1.x - v2.x, v1.y - v2.y);
}

inline VECTOR operator+(VECTOR const &v1, VECTOR const &v2)
{
	return VECTOR(v1.x + v2.x, v1.y + v2.y);
}

inline VECTOR operator*(VECTOR const &v, double a)
{
	return VECTOR(a * v.x, a * v.y);
}

inline VECTOR operator*(double a, VECTOR const &v)
{
	return VECTOR(a * v.x, a * v.y);
}

inline VECTOR operator/(VECTOR const &v, double a)
{
	return VECTOR(v.x / a, v.y / a);
}

inline VECTOR &VECTOR::operator+=(VECTOR const &v)
{
	x += v.x;
	y += v.y;
	return *this;
}

inline VECTOR &VECTOR::operator-=(VECTOR const &v)
{
	x -= v.x;
	y -= v.y;
	return *this;
}

inline int VECTOR::operator==(VECTOR const &v)
{
	if (x == v.x && y == v.y)
		return (1);
	else
		return (0);
}

inline int VECTOR::operator<(VECTOR const &v)
{
	if (x < v.x && y < v.y)
		return (1);
	else
		return (0);
}
inline int VECTOR::operator>(VECTOR const &v)
{
	if (x > v.x && y > v.y)
		return (1);
	else
		return (0);
}

inline double Dot (VECTOR const &v1, VECTOR const &v2)
{
	return (v1.x * v2.x + v1.y * v2.y);
}

inline double SDot (double x1, double y1, double x2, double y2)
{
	return (x1 * x2 + y1 * y2);
}

inline double PerpDot (VECTOR const &v1, VECTOR const &v2)
{
	return (v1.y * v2.x - v1.x * v2.y);
}

inline double SPerpDot (double x1, double y1, double x2, double y2)
{
	return (y1 * x2 - x1 * y2);
}

inline double Length (VECTOR const &v)
{
	return (sqrt (v.x * v.x + v.y * v.y));
}

inline double SLength (double x, double y)
{
	return (sqrt (x * x + y * y));
}

inline VECTOR Normal (VECTOR const &v)
{
	double oneOverLength = 1 / Length (v);
	return VECTOR (oneOverLength * v.x, oneOverLength * v.y);
}

inline VECTOR Perp (VECTOR const &v)
{
	return VECTOR (-v.y, v.x);
}

inline vectorStruct::vectorStruct () : x(0), y(0)
{
}

inline vectorStruct::vectorStruct (double newX, double newY) : x(newX), y(newY)
{
}
