/*
 * atanks - obliterate each other with oversize weapons
 * Copyright (C) 2003  Thomas Hudson
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * */

#include "globaldata.h"
#include "collision.h"

COLLISION_MESH::COLLISION_MESH (GLOBALDATA *global, ENVIRONMENT *env)
{
	_global = global;
	setupCollisionPoints ();
}
COLLISION_MESH::COLLISION_MESH (GLOBALDATA *global, ENVIRONMENT *env, int radius)
{
	_global = global;
	setupCollisionPoints (radius);
}
COLLISION_MESH::COLLISION_MESH (GLOBALDATA *global, ENVIRONMENT *env, BITMAP *bitmap)
{
	_global = global;
}

VECTOR COLLISION_MESH::setupCollisionPoints ()
{
	centreOfMass = VECTOR(0,0);
	points = new VECTOR(0,0);
	numPoints = 1;
	normals = NULL;

	return (centreOfMass);
}
VECTOR COLLISION_MESH::setupCollisionPoints (int radius)
{
	centreOfMass = VECTOR(0,0);
	numPoints = (int)(2 * PI * radius);
	double spokeStep = 360 / numPoints;
	points = new VECTOR[numPoints];
	normals = new VECTOR[numPoints];
	for (double spokeCount = 0; spokeCount < numPoints; spokeCount ++) {
		//// HERE right-angles?
		int spokeAngle = (int)(spokeCount * spokeStep);
		VECTOR pointAngle = _global->slope[spokeAngle];
		points[(int)spokeCount] = pointAngle * radius;
		normals[(int)spokeCount] = pointAngle;
	}

	return (centreOfMass);
}
VECTOR COLLISION_MESH::setupCollisionPoints (BITMAP *bitmap)
{
	centreOfMass = calcCentreOfMass (bitmap);
	points = new VECTOR[numPoints];
	normals = new VECTOR[numPoints];
	VECTOR scanPos = VECTOR (0,0);
	while (getpixel (bitmap, (int)scanPos.x, (int)scanPos.y) == PINK) {
		scanPos.x++;
		if (scanPos.x >= bitmap->w) {
			scanPos.x = 0;
			scanPos.y++;
			if (scanPos.y >= bitmap->h) {
				break;
			}
		}
	}
	if (scanPos.y < bitmap->h) {
		//HERE map edges/generate points
		//VECTOR

	}

	return (centreOfMass);
}

VECTOR COLLISION_MESH::calcCentreOfMass (BITMAP *bitmap)
{
	static VECTOR totalLocationOfMass = VECTOR (0,0);
	int massCount = 0;
	for (int x = 0; x < bitmap->w; x++)
		for (int y = 0; y < bitmap->h; y++)
			if (getpixel (bitmap, x, y) != PINK) {
				totalLocationOfMass += VECTOR(x,y);
				massCount++;
			}
	totalLocationOfMass = totalLocationOfMass / massCount;

	return (totalLocationOfMass);
}

int COLLISION_MESH::pointCollisionTest (VECTOR point, BITMAP *bitmap, VECTOR location, int rotation)
{
	VECTOR newPoint = VECTOR(PerpDot (_global->slope[rotation], point),
				Dot (_global->slope[rotation], point));
	//double newx = x * cos (angle) - y * sin (angle)
	//double newy = x * sin (angle) + y * cos (angle)
	newPoint += location;
	if (getpixel (bitmap, (int)newPoint.x, (int)newPoint.y) != PINK)
		return (1);
	else
		return (0);
}

VECTOR	COLLISION_MESH::collisionTest (BITMAP *bitmap, VECTOR location, int rotation)
{
	if (numPoints == 1 && points[0] == VECTOR(0,0)) {
		if (getpixel (bitmap, (int)location.x, (int)location.y) != PINK)
			return (VECTOR(0,0));
		else
			return (INVALID_VECTOR);
	}
	for (int pointCount = 0; pointCount < numPoints; pointCount++) {
		if (pointCollisionTest (points[pointCount], bitmap, location, rotation)) {
			return (points[pointCount]);
		}
	}
	return (INVALID_VECTOR);
}
