#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "main.h"
#include "parser.h"

ENUM_CONVERSION weaponTypeEnumConversion[] = {
	{"Beam", WEAPONTYPE_BEAM},
	{"Ballistic", WEAPONTYPE_BALLISTIC},
};

ENUM_CONVERSION triggerEnumConversion[] = {
	{"None", TRIGGER_NONE},
	{"Instant", TRIGGER_INSTANT},
	{"Collision", TRIGGER_COLLISION},
	{"Resurface", TRIGGER_RESURFACE},
	{"Zenith", TRIGGER_ZENITH},
	{"ComeToRest", TRIGGER_COMETOREST},
	{NULL, 0}
};

ENUM_CONVERSION dispersionEnumConversion[] = {
	{"None", DISPERSE_NONE},
	{"Spherical", DISPERSE_SPHERICAL},
	{"Parallel", DISPERSE_PARALLEL},
	{NULL, 0}
};

ENUM_CONVERSION disperseDirectionEnumConversion[] = {
	{"None", DISPERSEDIR_NONE},
	{"Up", DISPERSEDIR_UP},
	{"Down", DISPERSEDIR_DOWN},
	{"Forwards", DISPERSEDIR_FORWARDS},
	{"Reverse", DISPERSEDIR_REVERSE},
	{NULL, 0}
};

ENUM_CONVERSION detonationEnumConversion[] = {
	{"None", DETONATIONTYPE_NONE},
	{"Explosion", DETONATIONTYPE_EXPLOSION},
	{"Shaped Horizontal", DETONATIONTYPE_SHAPEDHORIZONTAL},
	{"Erase Wedge", DETONATIONTYPE_ERASEWEDGE},
	{"Earthquake", DETONATIONTYPE_EARTHQUAKE},
	{"Slow Burn", DETONATIONTYPE_SLOWBURN},
	{"Erase", DETONATIONTYPE_ERASE},
	{"Dirt", DETONATIONTYPE_DIRT},
};

ENUM_CONVERSION physicsTypeEnumConversion[] = {
	{"Normal", PHYSICSTYPE_NORMAL},
	{"Roll", PHYSICSTYPE_ROLL},
	{"Electricity", PHYSICSTYPE_ELECTRICITY},
	{"Lazer", PHYSICSTYPE_LAZER},
};

ENUM_CONVERSION repelTypeEnumConversion[] = {
	{"Yes", REPELTYPE_YES},
	{"No", REPELTYPE_NO},
};
ENUM_CONVERSION angleTypeEnumConversion[] = {
	{"Forwards", ANGLETYPE_FORWARDS},
	{"Spin", ANGLETYPE_SPIN},
};
ENUM_CONVERSION gravityTypeEnumConversion[] = {
	{"Yes", GRAVITYTYPE_YES},
	{"No", GRAVITYTYPE_NO},
};
ENUM_CONVERSION subterraneanTypeEnumConversion[] = {
	{"No", SUBTERRANEANTYPE_NO},
	{"Sink", SUBTERRANEANTYPE_SINK},
	{"Float", SUBTERRANEANTYPE_FLOAT},
};

PARSER_TOKEN weaponParserTokens[] = {
	{"Name",	TOKENTYPE_STRING, offsetof (WEAPON, name)},
	{"Description",	TOKENTYPE_STRING, offsetof (WEAPON, description)},
	{"Cost",	TOKENTYPE_INT, offsetof (WEAPON, cost)},
	{"BundlesOf",	TOKENTYPE_INT, offsetof (WEAPON, amt)},
	{"WeaponType",	TOKENTYPE_ENUM, offsetof (WEAPON, weaponType), 0, weaponTypeEnumConversion},
	{"DetonationType",	TOKENTYPE_ENUM, offsetof (WEAPON, detonationType), 0, detonationEnumConversion},
	{"PhysicsType",	TOKENTYPE_ENUM, offsetof (WEAPON, physicsType), 0, physicsTypeEnumConversion},
	{"DispersionType",	TOKENTYPE_ENUM, offsetof (WEAPON, dispersionType), 0, dispersionEnumConversion},
	{"DisperseDirection",	TOKENTYPE_ENUM, offsetof (WEAPON, disperseDirection), 0, disperseDirectionEnumConversion},
	{"Repel",	TOKENTYPE_ENUM, offsetof (WEAPON, repelType), 0, repelTypeEnumConversion},
	{"Gravity",	TOKENTYPE_ENUM, offsetof (WEAPON, gravityType), 0, gravityTypeEnumConversion},
	{"Orientation",	TOKENTYPE_ENUM, offsetof (WEAPON, angleType), 0, angleTypeEnumConversion},
	{"Subterranean",	TOKENTYPE_ENUM, offsetof (WEAPON, subterraneanType), 0, subterraneanTypeEnumConversion},
	{"Mass",	TOKENTYPE_DOUBLE, offsetof (WEAPON, mass)},
	{"Drag",	TOKENTYPE_DOUBLE, offsetof (WEAPON, drag)},
	{"ExplosionRadius",	TOKENTYPE_INT, offsetof (WEAPON, radius)},
	{"ObjectRadius",TOKENTYPE_DOUBLE, offsetof (WEAPON, objRadius)},
	{"ComeToRestPeriod",	TOKENTYPE_DOUBLE, offsetof (WEAPON, comeToRestPeriod)},
	{"ComeToRestArea",	TOKENTYPE_INT, offsetof (WEAPON, comeToRestArea)},
	{"Elasticity",	TOKENTYPE_DOUBLE, offsetof (WEAPON, elasticity)},
	{"Grip",	TOKENTYPE_DOUBLE, offsetof (WEAPON, grip)},
	{"SoundNumber",	TOKENTYPE_INT, offsetof (WEAPON, sound)},
	{"ETime",	TOKENTYPE_INT, offsetof (WEAPON, etime)},
	{"Damage",	TOKENTYPE_INT, offsetof (WEAPON, damage)},
	{"EFrames",	TOKENTYPE_INT, offsetof (WEAPON, eframes)},
	{"InGameImage",	TOKENTYPE_BITMAP, offsetof (WEAPON, inGameImage)},
	{"StockImage",	TOKENTYPE_BITMAP, offsetof (WEAPON, stockImage)},
	{"Spread",	TOKENTYPE_INT, offsetof (WEAPON, spread)},
	{"Delay",	TOKENTYPE_INT, offsetof (WEAPON, delay)},
	{"Trigger",	TOKENTYPE_ENUM, offsetof (WEAPON, trigger), 0, triggerEnumConversion},
	{"TechLevel",	TOKENTYPE_INT, offsetof (WEAPON, techLevel)},
	{"Warhead",	TOKENTYPE_INT, offsetof (WEAPON, warhead)},
	{"NumSubmunitions",	TOKENTYPE_INT, offsetof (WEAPON, numSubmunitions)},
	{"Submunition",	TOKENTYPE_STRING, offsetof (WEAPON, submunitionName)},
	{"ImpartVelocity",	TOKENTYPE_DOUBLE, offsetof (WEAPON, impartVelocity)},
	{"Divergence",	TOKENTYPE_INT, offsetof (WEAPON, divergence)},
	{"SpreadVariation",	TOKENTYPE_DOUBLE, offsetof (WEAPON, spreadVariation)},
	{"LaunchSpeed",	TOKENTYPE_DOUBLE, offsetof (WEAPON, launchSpeed)},
	{"SpeedVariation",	TOKENTYPE_DOUBLE, offsetof (WEAPON, speedVariation)},
	{"Countdown",	TOKENTYPE_INT, offsetof (WEAPON, countdown)},
	{"CountVariation",	TOKENTYPE_DOUBLE, offsetof (WEAPON, countVariation)},
	{"EndWeaponDefinition", TOKENTYPE_TERMINATOR},
	{"EndNaturalDefinition", TOKENTYPE_TERMINATOR},
	{NULL, TOKENTYPE_NONE, 0},
};

CONFIG_SECTION weaponConfigSection = {
	"Weapons", (void**)&weapon, 0, 0, sizeof (WEAPON), weaponParserTokens
};
CONFIG_SECTION naturalsConfigSection = {
	"Weapons", (void**)&natural, 0, 0, sizeof (WEAPON), weaponParserTokens
};


PARSER_TOKEN deviceParserTokens[] = {
	{"Name",	TOKENTYPE_STRING,	offsetof (ITEM, name)},
	{"Description",	TOKENTYPE_STRING,	offsetof (ITEM, description)},
	{"Cost",	TOKENTYPE_INT,		offsetof (ITEM, cost)},
	{"BundlesOf",	TOKENTYPE_INT,		offsetof (ITEM, amt)},
	{"Selectable",	TOKENTYPE_INT,		offsetof (ITEM, selectable)},
	{"TechLevel",	TOKENTYPE_INT,		offsetof (ITEM, techLevel)},
	{"Sound",	TOKENTYPE_INT,		offsetof (ITEM, sound)},
	{"Val1",	TOKENTYPE_DOUBLE,	offsetof (ITEM, vals[0])},
	{"Val2",	TOKENTYPE_DOUBLE,	offsetof (ITEM, vals[1])},
	{"Val3",	TOKENTYPE_DOUBLE,	offsetof (ITEM, vals[2])},
	{"Val4",	TOKENTYPE_DOUBLE,	offsetof (ITEM, vals[3])},
	{"EndDeviceDefinition", TOKENTYPE_TERMINATOR},
};

PARSER_TOKEN topLevelParserTokens[] = {
	{"WeaponDefinition",	TOKENTYPE_KEYWORD,	0, sizeof (WEAPON), NULL, &weaponConfigSection},
	{"NaturalDefinition",	TOKENTYPE_KEYWORD,	0, sizeof (WEAPON), NULL, &naturalsConfigSection},
//	{"DeviceDefinition",	TOKENTYPE_KEYWORD,	deviceParserTokens, sizeof (ITEM)}
//	{"EndConfig", TOKENTYPE_TERMINATOR},
};

CONFIG_SECTION topLevelConfigSection = {
	"Top Level", NULL, 0, 0, 0, topLevelParserTokens
};



int parseTokens (FILE *file, CONFIG_SECTION *config) {
	char tmpString[1024];
	PARSER_TOKEN *parserTokens = config->parserTokens;
	char *structure;

	if (config->dest) {
		int copyDefaults;
		if (!*config->dest) {
			*config->dest = malloc (config->structSize);
			config->destSize += config->structSize;
			copyDefaults = FALSE;
		} else {
			*config->dest = realloc (*config->dest, config->destSize + config->structSize);
			config->destSize += config->structSize;
			copyDefaults = TRUE;
		}
       		structure = &(((char*)(*config->dest))[config->destSize - config->structSize]);
		if (copyDefaults)
			memcpy (structure, *config->dest, config->structSize);
		config->numEntries++;
	}

	while (fgets (tmpString, 1024, file)) {
		char *value = strchr (tmpString, ':');
		PARSER_TOKEN *tokenEntry = NULL;
		if (!value) {
			continue;
		}
		char *newline = strchr (value, '\n');
		char *keyword = tmpString;

		if (newline)
			*newline = '\0';

		while (*keyword == '\t' || *keyword == ' ')
			keyword++;
		if (value)
			*value = '\0';
		value++;
		while (*value == '\t' || *value == ' ')
			value++;

		tokenEntry = parserTokens;
		while (tokenEntry->token != NULL && strcmp (keyword, tokenEntry->token))
			tokenEntry++;
		if (tokenEntry->token != NULL) {
			// token matches
			void *subStruct;
			char *dataDest;
			ENUM_CONVERSION *conversion;
			if (structure)
				dataDest = (char*)structure + tokenEntry->offset;
			else
				dataDest = NULL;

			switch (tokenEntry->type) {
				case TOKENTYPE_KEYWORD:
					subStruct = malloc (tokenEntry->structSize);
					parseTokens (file, tokenEntry->config);
					break;
				case TOKENTYPE_STRING:
					if (dataDest) {
						char *newString = strdup (value);
						*(char**)dataDest = newString;
					}
					break;
				case TOKENTYPE_INT:
					if (dataDest)
						*((int*)dataDest) = strtol (value, NULL, 10);
					break;
				case TOKENTYPE_DOUBLE:
					if (dataDest)
						*((double*)dataDest) = strtod (value, NULL);
					break;
				case TOKENTYPE_ENUM:
					if (dataDest) {
						conversion = tokenEntry->conversion;
						int enumCount;
						for (enumCount = 0; conversion[enumCount].define != NULL; enumCount++)
							if (!strcmp (conversion[enumCount].define, value))
								break;
						if (conversion[enumCount].define) {
							*((int*)dataDest) = conversion[enumCount].number;
						} else {
							printf ("%s is not a valid value\n", value);
							printf ("Valid tokens are:\n");
							enumCount = 0;
							while (conversion[enumCount].define != NULL) {
								printf ("\t%s\n", conversion[enumCount].define);
								enumCount++;
							}
						}
					}
					break;
				case TOKENTYPE_BITMAP:
					if (dataDest) {
						struct stat statbuf;
						if (!stat (value, &statbuf)) {
							*((BITMAP**)dataDest) = load_bitmap (value, NULL);
							if (!dataDest) {
								fprintf (stderr, "Failed to load bitmap %s\n", value);
								exit (1);
							}
						} else {
							char errbuf[1024];
							sprintf (errbuf, "Loading bitmap %s failed", value);
							perror (errbuf);
							exit (1);
						}
					}
					break;
				case TOKENTYPE_TERMINATOR:
					return (0);
					break;
				default:
					printf ("Unknown token type: %d\n", tokenEntry->type);
			}
		} else {
			printf ("Unknown token \"%s\", value \"%s\"\n",
				keyword, value);
		}
	}
	return (0);
}

