/*
 * atanks - obliterate each other with oversize weapons
 * Copyright (C) 2003  Thomas Hudson
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * */


#include "main.h"
#include "physobj.h"
#include "globaldata.h"
#include "environment.h"


int PHYSICAL_OBJECT::applyPhysics ()
{
	PHYSICAL_OBJECT *ltank;
	/*if (x + xv < 1 || x + xv > (_global->screenWidth-1)) {
		xv = -xv;	//bounce on the border
	} else {
		//double accel = (_env->wind - xv) / CALC_DRAG_ACCEL(_global->identicalTrajectories, mass, drag) * _env->viscosity;
		double accel = calcWindAccel ((int)_global->identicalTrajectories, _env->wind - xv, mass, drag, _env->viscosity);
		xv += accel;
		x += xv;
	}
	if (y + yv >= _global->screenHeight) {
		double startX, startY;

		if (checkPixelsBetweenTwoPoints (_global, _env, &startX, &startY, x+xv, y+yv)) {
			hitSomething = 1;
		} else {
			yv = -yv * 0.5;
			xv *= 0.95;
			if (fabs(xv) + fabs(yv) < 0.8)
				hitSomething = 1;
		}
	}
	y += yv;
	yv += _env->gravity * (100.0 / FRAMES_PER_SECOND);
	*/

	if (angleType == ANGLETYPE_FORWARDS) {
		angle = (atan (yv / xv) * (180 / PI) * ACHANGE) - 64 + ((xv < 0) ? 128 : 0);
	} else if (angleType == ANGLETYPE_SPIN) {
		angle += (int)spin;
		while (angle > 360)
			angle -= 360;
		while (angle < 0)
			angle += 360;
	}

	VECTOR accel = VECTOR (0,0);
	for (int objCount = 0; (ltank = (PHYSICAL_OBJECT*)_env->getNextOfClass (TANK_CLASS, &objCount)) && ltank; objCount++) {
		if (ltank->player != player) {
			double xaccel = 0, yaccel = 0;
			ltank->repulse (x, y, &xaccel, &yaccel);
			accel.x += xaccel;
			accel.y += yaccel;
		}
	}

	double distanceRemaining = sqrt ((xv * xv) + (yv * yv));
	double fullDistance = distanceRemaining;
	VECTOR step = VECTOR (xv, yv) / fullDistance;
			
	printf ("%d\n", physicsType);
	if (physicsType == PHYSICSTYPE_NORMAL) {
		detonate = checkPixelsBetweenPrevAndNow ();
		/*if (rand () % 5 == 0) {
			DECOR *decor;
			decor = new DECOR (_global, _env, x, y, xv, yv, expSize / 20, DECOR_SMOKE);
			if (!decor) {
				perror ("missile.cc: Failed allocating memory for decor in PHYSOBJ::applyPhysics");
				exit (1);
			}
		}*/
	} else if (physicsType == PHYSICSTYPE_ROLL) {
		int spokeStep = (int)(360 / (2 * PI * objRadius));
		// spokeStart is the direction of travel - 90 degrees
		int directionOfTravel = (int)(atan2 (yv, xv) / PI * 180);
		while (directionOfTravel < 0)
			directionOfTravel += 360;
		int spokeStart = directionOfTravel + 270;
		if (spokeStart > 360)
			spokeStart -= 360;
		if (spokeStep < 1)
			spokeStep = 1;
		//int applyGravity = TRUE;
		int hitAngle = spokeStart;
		int spokeStop = spokeStart + 180;
		while (distanceRemaining > 0 && fullDistance > 0) {
			if (hitAngle < 0)
				hitAngle = spokeStart;
			hitAngle = scanLeadingArc (hitAngle, spokeStop, spokeStep);

			if (hitAngle >= 0 && hitAngle < 360) {
				// point of collision normal
				VECTOR pcn = _global->slope[hitAngle];
				//pcn.x = _global->slope[hitAngle][1];
				//pcn.y = _global->slope[hitAngle][0];

				// collision point
				VECTOR cp = pcn * objRadius;

				// Velocity perpendicular to radius
				VECTOR pv = Perp (cp) * spin;

				// collision normal taking
				//   spin into account
				//VECTOR cn = Normal (cp - (pv * grip));
				VECTOR cn = pcn;

				// velocity normal
				VECTOR vn = _global->slope[directionOfTravel];
				//vn.x = _global->slope[directionOfTravel][1];
				//vn.y = _global->slope[directionOfTravel][0];

				VECTOR combinedV (xv + pv.x, yv + pv.y);
				VECTOR combinedVNorm;
				combinedVNorm = Normal (combinedV);

				// Is it moving towards the
				//   point of collision?
				if (Dot (cn, combinedVNorm) > 0) {
					if (distanceRemaining < 1) {
						x -= step.x * distanceRemaining;
						y -= step.y * distanceRemaining;
					} else {
						x -= step.x;
						y -= step.y;
					}
					//double perpdot = PerpDot (cp, cn);
					double oneOverMass = 1 / mass;
					double oneOverInertia = 1 / inertia;
					double impulseNumerator = -(1 + elasticity) * Dot (combinedV, cn);
					double impulseDenominator = oneOverMass;// + oneOverInertia * perpdot * perpdot;
					double impulse = impulseNumerator / impulseDenominator;

					xv += impulse * cn.x * oneOverMass;
					yv += impulse * cn.y * oneOverMass;
					double perpdotV = PerpDot (combinedV, cn);
					spin += impulse * oneOverInertia * perpdotV;


					// Calc new settings
					directionOfTravel = (int)(atan2 (yv, xv) / PI * 180);
					if (directionOfTravel < 0)
						directionOfTravel += 360;
					spokeStart = directionOfTravel + 270;
					if (spokeStart > 360)
						spokeStart -= 360;
					double speed = SLength (xv, yv);
					distanceRemaining *= (speed / fullDistance);
					fullDistance = speed;
					step = VECTOR (xv, yv) / fullDistance;
					if (fullDistance < 0.1)
						detonate = 1;
					//applyGravity = FALSE;
				} else {
					hitAngle += spokeStep;
					continue;
				}
			}
			if (distanceRemaining < 1) {
				x += step.x * distanceRemaining;
				y += step.y * distanceRemaining;
			} else {
				x += step.x;
				y += step.y;
			}
			distanceRemaining--;
		}

		//PHYSICAL_OBJECT::applyPhysics ();
		//x += xv;
		//y += yv;
		//if (applyGravity)
		//	yv += _env->gravity;

		//angle += spin / PI * 180;

		//while (angle > 360)
		//	angle -= 360;
		//while (angle < 0)
		//	angle += 360;

	}

	if (gravityType == GRAVITYTYPE_YES) {

		if ((subterraneanType == SUBTERRANEANTYPE_NO) || getpixel (_env->terrain, (int)x, (int)y) == PINK) {
			accel.y += _env->gravity * (100.0 / FRAMES_PER_SECOND);
		} else if (subterraneanType == SUBTERRANEANTYPE_FLOAT) {
			accel.y -= _env->gravity * 0.05 * (100.0 / FRAMES_PER_SECOND);
		} else if (subterraneanType == SUBTERRANEANTYPE_SINK) {
			accel.y += _env->gravity * 0.05 * (100.0 / FRAMES_PER_SECOND);
		}
	}

	if (x + xv < 1 || x + xv > (_global->screenWidth-1)) {
		xv = -xv;	//bounce on the border
	} else {
		x += xv;
	}

	//hit screen bottom then change the y velocity direction
	if (y + yv >= _global->screenHeight) {
		double startX, startY;

		if (checkPixelsBetweenTwoPoints (_global, _env, &startX, &startY, x+xv, y+yv)) {
			hitSomething = 1;
		} else {
			yv = -yv * 0.5;
			xv *= 0.95;
		if (fabs(xv) + fabs(yv) < 0.8)
			hitSomething = 1;
		}
	} else if (y + yv < 0) {
		yv = -yv *0.5;
		xv *= 0.95;
	}

	xv += accel.x;
	yv += accel.y;
	x += xv;
	y += yv;

	return (hitSomething);
}

void PHYSICAL_OBJECT::repulse (double xpos, double ypos, double *xaccel, double *yaccel)
{
	if (repulsion != 0) {
		double xdist = xpos - x;
		double ydist = ypos - y;
		double distance2 = (xdist * xdist) + (ydist * ydist);
		double distance = sqrt (distance2);

		if (distance < 60 + sqrt ((double)repulsion)) {
			*xaccel = repulsion * (xdist / distance) / distance2;
			*yaccel = repulsion * (ydist / distance) / distance2;
		}
	}
}


int PHYSICAL_OBJECT::checkPixelsBetweenPrevAndNow ()
{
	double startX = x - xv;
	double startY = y - yv;

	int leftOver = checkPixelsBetweenTwoPoints (_global, _env, &startX, &startY, x, y);
	if (leftOver) {
		x = startX;
		y = startY;
		return (leftOver);
	}

	return (0);
}

int PHYSICAL_OBJECT::scanLeadingArc (int spokeStart, int spokeStop, int spokeStep)
{
	for (int spokeCount = 0; spokeStart + spokeCount < spokeStop; spokeCount += spokeStep) {
		int spokeAngle = (spokeStart + spokeCount) % 360;
		// HERE right-angles?
		VECTOR spokePos = VECTOR (x, y) + (_global->slope[spokeAngle] * objRadius);
		//double spokeXPos = x + _global->slope[spokeAngle][1] * objRadius;
		//double spokeYPos = y + _global->slope[spokeAngle][0] * objRadius;
		//printf ("%d,%d,%d\n", spokeStart, spokeAngle, spokeStop);
		//putpixel (screen, (int)(spokeXPos + 30), (int)(spokeYPos + 30), BLACK);
		if (spokePos.x < 0 ||
			spokePos.x >= _global->screenWidth || 
			spokePos.y < 0 ||
			spokePos.y > _global->screenHeight)
			continue;
		if (getpixel (_env->terrain, (int)spokePos.x, (int)spokePos.y) != PINK)
			return (spokeAngle);
	}

	return (-1); // Hit something if 0-359
}

int checkPixelsBetweenTwoPoints (GLOBALDATA *global, ENVIRONMENT *env, double *startX, double *startY, double endX, double endY)
{
	int landFound = 0;

	// only check if on screen
	double xDist = endX - *startX;
	double yDist = endY - *startY;
	double length;
	double xInc;
	double yInc;

	if (xDist < 2 && yDist < 2) {
		if (endX > 0 && endX < global->screenWidth &&
			endY < global->screenHeight && endY > 0 &&
			(getpixel (env->terrain, (int)endX, (int)endY) != PINK))
			return (1);
		else
			return (0);
	}

       	length = sqrt ((xDist * xDist) + (yDist * yDist));
	yInc = yDist / length;
	xInc = xDist / length;

	// sanity check
	if (length > global->screenWidth + global->screenHeight)
		length = global->screenWidth + global->screenHeight;

	//check all pixels along line for land
	int lengthPos;
	for (lengthPos = 0; lengthPos < length; lengthPos++) {
		//found land
		if (*startX > 0 && *startX < global->screenWidth && *startY < global->screenHeight && *startY > 0 && (getpixel (env->terrain, (int)*startX, (int)*startY) != PINK)) {
			// Leaves startX and Y at current point
			landFound = 1;
			break;
		}
		*startX += xInc;
		*startY += yInc;
	}

	return ((int)(length - lengthPos));
}

double calcWindAccel (int ident, double wind, double mass, double drag, double visc) {
	return ((ident)? wind * 0.01 : wind / mass * drag * visc);
}


void PHYSICAL_OBJECT::initialise ()
{
	VIRTUAL_OBJECT::initialise ();
	hitSomething = 0;
}

/*int PHYSICAL_OBJECT::isSubClass (int classNum)
{
	if (classNum != PHYSICAL_OBJECT_CLASS)
		return (VIRTUAL_OBJECT::isSubClass (classNum));
	else
		return (TRUE);
}*/
